import typing

import dataclasses
import logging
import os
import sys

from typing import Any, Dict, Tuple, Union

from polyswarmdconfig.exceptions import MissingConfigValueError


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Config:
    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> 'Config':
        try:
            return cls(**cls.populate(config))
        except TypeError as e:
            logger.exception('Missing value while populating %s', cls.__name__)
            raise MissingConfigValueError from e

    @classmethod
    def from_dict_and_environment(cls, config: Dict[str, Any]) -> 'Config':
        cls.overlay_environment(config)
        try:
            return cls(**cls.populate(config))
        except TypeError as e:
            logger.exception('Missing value while populating %s', cls.__name__)
            raise MissingConfigValueError from e

    @classmethod
    def populate(cls, config: Dict[str, Any]) -> Dict[str, Any]:
        module = sys.modules[cls.__module__]
        return {k: cls.get_value(k, v, module) for k, v in config.items() if typing.get_type_hints(cls).get(k)}

    @classmethod
    def get_value(cls, key, value, module):
        if not isinstance(value, dict):
            return cls.correct_type(key, value)
        elif module and hasattr(module, key.capitalize()):
            sub_config_class = getattr(module, key.capitalize())
            if issubclass(sub_config_class, Config):
                sub_config = sub_config_class.from_dict(value)
                return sub_config

    @classmethod
    def correct_type(cls, key: str, value: Any) -> Any:
        cast = typing.get_type_hints(cls).get(key)
        if cast and cast in [int, str]:
            return cast(value)
        elif cast and cast in [bool]:
            return cls._attempt_to_extract_bool(value)
        elif cast and cast in [list]:
            return cls._attempt_to_extract_list(value)
        else:
            return value

    @classmethod
    def overlay_environment(cls, config: Dict[str, Any]):
        name = cls.__name__.upper()
        for key, value in os.environ.items():
            if key.startswith(name):
                cls.overlay_matching_value(key.replace(f'{name}_', ''), value, config)

    @classmethod
    def overlay_matching_value(cls, key: str, value: Union[str, int, bytes], config: Dict[str, Any]):
        module = sys.modules[cls.__module__]
        current = config
        rest = key
        while True:
            title, rest = Config.split(rest)
            if module and hasattr(module, title.capitalize()):
                module = sys.modules[getattr(module, title.capitalize()).__module__]
                found_value = current.get(title.lower())
                if found_value and isinstance(found_value, dict):
                    current = found_value
                else:
                    current[title] = {}
                    current = current[title]
            else:
                # Not a sub config, just store the value
                current['_'.join([title, rest.lower()]) if rest else title] = value
                break

    @staticmethod
    def split(key: str) -> Tuple[str, str]:
        separated = key.split('_', 1)
        title = separated[0].lower()
        rest = separated[1] if len(separated) > 1 else ''
        return title, rest

    @staticmethod
    def _attempt_to_extract_bool(value):
        if isinstance(value, str):
            return value.lower() not in ('false', '0', '')
        else:
            return bool(value)

    @staticmethod
    def _attempt_to_extract_list(value):
        items_delimiter = ','
        return list(map(str.strip, value.split(items_delimiter)))



import dataclasses

from urllib.parse import urlparse
from consul import Consul as ConsulClient
from typing import Optional

from polyswarmdconfig.config import Config


@dataclasses.dataclass
class Consul(Config):
    uri: str
    token: Optional[str] = None
    client: ConsulClient = dataclasses.field(init=False)

    def __post_init__(self):
        u = urlparse(self.uri)
        self.client = ConsulClient(host=u.hostname, port=u.port, scheme=u.scheme, token=self.token)



import dataclasses

from redis import Redis as RedisClient
from typing import Optional

from polyswarmdconfig.config import Config


@dataclasses.dataclass
class Redis(Config):
    uri: Optional[str] = None
    client: Optional[RedisClient] = dataclasses.field(init=False, default=None)

    def __post_init__(self):

        if self.uri:
            self.client = RedisClient.from_url(self.uri)



from abc import ABC, abstractmethod


class AbstractArtifactServiceClient(ABC):
    """
    Abstract class that defines the interface all Artifact Service Clients must follow
    These functions are the features that polyswarmd requires in some artifact storage
    Namely, checking the status of the service, adding and retrieving artifacts, wrapping artifacts in some
    logical grouping (directory-like) and finding the details
    """
    name: str
    reachable_endpoint: str

    def __init__(self, name, reachable_endpoint):
        self.name = name
        self.reachable_endpoint = reachable_endpoint

    @abstractmethod
    def add_artifacts(self, artifacts, session):
        """
        Add a list of artifacts to the service

        :param artifacts: list[tuple]: List of files to be added to the service.
        :param session: connection session
        :return: (str) URI for the directory of files
        :raises ArtifactServiceException: If service is unreachable or gives non-200 response
        """
        raise NotImplementedError()

    @abstractmethod
    def add_artifact(self, artifact, session, redis=None):
        """
        Add a list of artifacts to the service

        :param artifact: List of files to be added to the service.
        :param session: connection session
        :param redis: redis connection for caching contents
        :return: (str) URI for the added file
        :raises ArtifactServiceException: If service is unreachable or gives non-200 response
        """
        raise NotImplementedError()

    def check_uri(self, uri):
        """
        Check if the given uri is valid for the service.
        :param uri: uri to check
        :return: bool
        :raises ArtifactServiceException: If service is unreachable or gives non-200 response
        """
        raise NotImplementedError()

    @abstractmethod
    def details(self, identifier, index, session):
        """
        Get the details for a particular artifact with the given identifier.

        :param identifier: uri or other identifier for the artifact
        :param index: index of the artifact in a directory, or 0 if file is not a directory
        :param session: connection session
        :return: (dict) Dict of values about the file in question
        :raises ArtifactServiceException: If service is unreachable or gives non-200 response
        """
        raise NotImplementedError()

    @abstractmethod
    def get_artifact(self, identifier, index, session, max_size=None, redis=None, as_url=False) -> bytes:
        """
        Get the artifact at the given identifer and index

        :param identifier: uri or other identifier for the artifact
        :param index: index of the artifact in a directory, or 0 if file is not a directory
        :param session: connection session
        :param max_size: optional max size of files to get
        :param redis: redis connection for reading from cache contents
        :param as_url: content will be an URL containing the payload, instead of the payload directly
        :return: (bytes) Byte content of the given file
        :raises ArtifactServiceException: If service is unreachable or gives non-200 response
        """
        raise NotImplementedError()

    @abstractmethod
    def ls(self, identifier, session):
        """
        List the files in the directory at the given identifier

        :param identifier: uri or other identifier for the directory (or directory-like object)
        :param session: connection session
        :return: (list[(string, string)]) List of tuples containing name, uri pairs
        :raises ArtifactServiceException: If service is unreachable or gives non-200 response
        """
        raise NotImplementedError()

    @abstractmethod
    def status(self, session):
        """
        Performs an request against an artifact service to determine current running status

        :param session: connection session
        :return: (dict) with key `online`
        :raises ArtifactServiceException: If service is unreachable or gives non-200 response
        """
        raise NotImplementedError()



import importlib
from typing import Any, List

import dataclasses
from polyswarmdconfig.artifactclient import AbstractArtifactServiceClient
from polyswarmdconfig.config import Config


DEFAULT_FALLBACK_SIZE = 10 * 1024 * 1024


class ClassModuleLoader:
    module_name: str
    class_name: str

    def __init__(self, module_name: str, class_name: str):
        self.module_name = module_name
        self.class_name = class_name

    def load(self):
        client_module = importlib.import_module(self.module_name)
        return getattr(client_module, self.class_name)


@dataclasses.dataclass
class Library(Config):
    module: str
    class_name: str
    args: List[Any] = dataclasses.field(default_factory=list)
    client: AbstractArtifactServiceClient = dataclasses.field(init=False)

    def __post_init__(self):
        self.client = ClassModuleLoader(self.module, self.class_name).load()(*self.args)


@dataclasses.dataclass
class Artifact(Config):
    library: Library
    limit: int = 256
    fallback_max_size: int = DEFAULT_FALLBACK_SIZE
    max_size: int = DEFAULT_FALLBACK_SIZE
    download_as_url: bool = True

    def __post_init__(self):
        if self.limit < 1 or self.limit > 256:
            raise ValueError(
                'Artifact limit must be greater than 0 and cannot exceed contract limit of 256'
            )

        if self.fallback_max_size < 1:
            raise ValueError('Fall back max artifact size must be above 0')

    @property
    def client(self):
        return self.library.client



from setuptools import find_packages, setup


setup(
    name='polyswarmd-config',
    version='0.1.0',
    description='Config module the PolySwarm Daemon',
    author='PolySwarm Developers',
    author_email='info@polyswarm.io',
    url='https://github.com/polyswarm/polyswarmd-config',
    license='MIT',
    install_requires=[
        'dataclasses==0.7; python_version=="3.6"',
        'python-consul==1.1.0',
        'redis==3.4.1',
    ],
    include_package_data=True,
    packages=find_packages('src'),
    package_dir={'': 'src/'},
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ]
)
